#
# ObjectFactories
#
# Default object factories
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public
# License (MPL)
#

"""
$Id: ObjectFactories.py,v 1.7 2003/05/21 00:03:56 philipp Exp $
"""

from interfaces.ObjectFactories import IFactoryRegistry, IObjectFactory

class FactoryRegistry:
    """
    Registry for object factories
    """

    __implements__ = IFactoryRegistry

    def __init__(self):
        self._reg = {}

    def register(self, meta_type, obj):
        self._reg[meta_type] = obj

    def getFactory(self, meta_type):
        return self._reg.get(meta_type, None)

    def getSupportedMetaTypes(self):
        return tuple(self._reg.keys())

factoryRegistry = FactoryRegistry()

#
# Available factories:
#
# DTML Method
# DTML Document
# Page Template
# File
# Script (Python)
# External Method
# Z SQL Method
# Parsed XML
#

class DTMLMethodFactory:

    __implements__ = IObjectFactory

    def createObject(self, id, title, data, **kw):
        from OFS.DTMLMethod import DTMLMethod
        resultObj = DTMLMethod(data, __name__=id)
        resultObj.id = id
        resultObj.title = title
        return resultObj

factoryRegistry.register("DTML Method", DTMLMethodFactory())

class DTMLDocumentFactory:

    __implements__ = IObjectFactory

    def createObject(self, id, title, data, **kw):
        from OFS.DTMLDocument import DTMLDocument
        resultObj = DTMLDocument(data, __name__=id)
        resultObj.id = id
        resultObj.title = title
        return resultObj

factoryRegistry.register("DTML Document", DTMLDocumentFactory())

class ZPTFactory:

    __implements__ = IObjectFactory

    def createObject(self, id, title, data, **kw):
        from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
        content_type = kw.get('content_type', 'text/html')
        resultObj = ZopePageTemplate(id, data, content_type)
        resultObj.pt_setTitle(title)
        return resultObj

factoryRegistry.register("Page Template", ZPTFactory())

class FileFactory:

    __implements__ = IObjectFactory

    def createObject(self, id, title, data, **kw):
        from OFS.Image import File
        content_type = kw.get('content_type', None)
        resultObj = File(id, title, data, content_type)
        return resultObj

factoryRegistry.register("File", FileFactory())

class ImageFactory:

    __implements__ = IObjectFactory

    def createObject(self, id, title, data, **kw):
        from OFS.Image import Image
        content_type = kw.get('content_type', None)
        resultObj = Image(id, title, data, content_type)
        return resultObj

factoryRegistry.register("Image", ImageFactory())

class PythonScriptFactory:
    
    __implements__ = IObjectFactory

    def createObject(self, id, title, data, **kw):
        from Products.PythonScripts.PythonScript import PythonScript
        resultObj = PythonScript(id)
        resultObj.write(data)
        resultObj.title = title
        return resultObj
    
factoryRegistry.register("Script (Python)", PythonScriptFactory())

class ExternalMethodFactory:

    __implements__ = IObjectFactory

    def createObject(self, id, title, data, module, function, **kw):
        from Products.ExternalMethod.ExternalMethod import ExternalMethod
        resultObj = ExternalMethod(id, title, module, function)
        return resultObj

factoryRegistry.register("External Method", ExternalMethodFactory())

class ZSQLMethodFactory:

    __implements__ = IObjectFactory

    def createObject(self, id, title, data, **kw):
        from Products.ZSQLMethods.SQL import SQL
        connection_id = kw.get('conntection_id', '')
        arguments = kw.get('arguments', '')
        resultObj = SQL(id, title, connection_id, arguments, data)
        return resultObj

factoryRegistry.register("Z SQL Method", ZSQLMethodFactory())

class ParsedXMLFactory:
    
    __implements__ = IObjectFactory

    def createObject(self, id, title, data, **kw):
        from Products.ParsedXML.ParsedXML import ParsedXML
        content_type = kw.get('content_type', 'text/xml')
        resultObj = ParsedXML(id, data, contentType=content_type)
        return resultObj

try:
    import Products.ParsedXML
    factoryRegistry.register("Parsed XML", ParsedXMLFactory())
except ImportError:
    pass
