#
# Properties
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public
# License (MPL)
#

"""
$Id: Properties.py,v 1.2 2003/05/21 00:03:56 philipp Exp $
"""

import types
from ZPublisher.Converters import type_converters

from interfaces.Properties import IPropertySerializer, \
     IPropertyAdapter, IPropertyAdapterRegistry
from util import escape_string, unescape_string

class PropertySerializer:
    """
    Property serializer

    Serializes properties into a string with one line per property
    containing::

      id:type=value
    """

    __implements__ = IPropertySerializer

    def serialize(self, properties):
        s = ''
        for prop in properties:
            value = str(prop['value'])
            value = escape_string(value)
            s += "%s:%s=%s\n" % (prop['id'], prop['type'], value)
        return s

    def deserialize(self, serials):
        assert type(serials) is type('')
        properties = []
        lines = serials.split('\n')
        for line in lines:
            if line.startswith('#'):
                continue
            try:
                key, value = line.split('=')
            except ValueError:
                continue
            id, typ = key.split(':')
            id.strip()
            typ.strip()
            value = unescape_string(value)
            if type_converters.has_key(typ):
                value = type_converters[typ](value)
            properties.append({
                'id':    id,
                'type':  typ,
                'value': value
                })
        return tuple(properties)

propertySerializer = PropertySerializer()

def getAllBases(klass):
    """
    Get all base classes for klass recursively. This works like
    __mro__ in python 2.2 and above.
    """
    bases = []
    for base in klass.__bases__:
        bases.append(base)
        bases += getAllBases(base)
    # remove doubles
    i = 0
    while i < len(bases):
        if bases[i] in bases[:i]:
            bases.pop(i)
        else:
            i += 1
    return bases

class PropertyAdapterRegistry:
    """
    Registry for property adapters
    """

    __implements__ = IPropertyAdapterRegistry

    def __init__(self):
        self._registry = {}

    def registerForClass(self, klass, adapter):
        self._registry[klass] = adapter

    def getAdapter(self, obj):
        registry = self._registry
        klass = obj.__class__
        if registry.has_key(klass):
            return registry[klass]
        for base in getAllBases(klass):
            if registry.has_key(base):
                return registry[base]

# make a singleton object
propertyAdapterRegistry = PropertyAdapterRegistry()

class OFSPropertyAdapter:
    """
    OFS property adapter
    """

    __implements__ = IPropertyAdapter

    def getProperties(self, obj):
        propertymap = obj.propertyMap()
        # propertyMap() returns copies of the actual definitions
        # so we can go a head and change stuff here
        for prop in propertymap:
            prop['value'] = getattr(obj, prop['id'])
            if prop.has_key('mode'):
                del prop['mode']
        return propertymap

    def setProperties(self, obj, properties):
        obj_props = {}
        # remember the original property dicts by their ids
        for prop in obj.propertyMap():
            obj_props[prop['id']] = prop
        for prop in properties:
            id = prop['id']
            # the property is already there
            if obj_props.has_key(id):
                # we have to be very careful not to replace a
                # dictionary and thus eliminate vital keys.
                # Therefore, we always make copies
                value = prop['value']
                prop = prop.copy()
                del prop['value']
                # we have a new property definition
                # (i.e. the type is different)
                if obj_props[id] != prop:
                    oldprop = obj_props[id].copy()
                    oldprop.update(prop)
                    prop = oldprop
                    obj_properties = list(obj._properties)
                    i = obj_properties.index(obj_props[id])
                    obj._properties = tuple(
                        obj_properties[:i] + [prop] + obj_properties[i+1:]
                        )
                obj._updateProperty(id, value)
                del obj_props[id]
            # the property is new, hence add it
            else:
                obj._setProperty(**prop)
        # delete the remaining; please note that this might fail
        for id in obj_props.keys():
            # we don't give manage_delProperties the whole sequence
            # of ids because it returns an error dialog if one
            # deletion fails. Thus, we have to call it for each id.
            obj.manage_delProperties([id])
        # P.S.: OFS.PropertyManager.PropertyManager is evil.

from OFS.PropertyManager import PropertyManager
propertyAdapterRegistry.registerForClass(
    PropertyManager,
    OFSPropertyAdapter()
    )

class SQLMethodPropertyAdapter:
    """
    SQL method property adapter
    """

    __implements__ = IPropertyAdapter

    def getProperties(self, obj):
        direct = int(not not obj.allow_simple_one_argument_traversal)
        return (
            {'id':'title',         'value':obj.title,         'type':'string'},
            {'id':'connection_id', 'value':obj.connection_id, 'type':'string'},
            {'id':'max_rows',      'value':obj.max_rows_,     'type':'int'},
            {'id':'max_cache',     'value':obj.max_cache_,    'type':'int'},
            {'id':'cache_time',    'value':obj.cache_time_,   'type':'int'},
            {'id':'class_name',    'value':obj.class_name_,   'type':'string'},
            {'id':'class_file',    'value':obj.class_file_,   'type':'string'},
            {'id':'class_name',    'value':obj.class_name_,   'type':'string'},
            {'id':'direct',        'value':direct,            'type':'int'}
            )

    def setProperties(self, obj, properties):
        props = {}
        for prop in properties:
            props[prop['id']] = prop['value']
        obj.title = props['title']
        obj.connection_id = props['connection_id']
        del props['title']
        del props['connection_id']
        obj.manage_advanced(**props)

from Products.ZSQLMethods.SQL import SQL
propertyAdapterRegistry.registerForClass(
    SQL,
    SQLMethodPropertyAdapter()
    )

class ExternalMethodPropertyAdapter:
    """
    Adapter for ExternalMethods
    """

    __implements__ = IPropertyAdapter

    def getProperties(self, obj):
        return (
            {'id':'title',    'value':obj.title,     'type':'string'},
            {'id':'module',   'value':obj._module,   'type':'string'},
            {'id':'function', 'value':obj._function, 'type':'string'},
            )

    def setProperties(self, obj, properties):
        props = {}
        for prop in properties:
            props[prop['id']] = prop['value']
        obj.manage_edit(**props)

from Products.ExternalMethod.ExternalMethod import ExternalMethod
propertyAdapterRegistry.registerForClass(
    ExternalMethod,
    ExternalMethodPropertyAdapter()
    )

type_map = {
    types.IntType:    'int',
    types.LongType:   'long',
    types.FloatType:  'float',
    types.StringType: 'string',
    }

class AttributePropertyAdapter:
    """
    Adapter that gathers property values from attributes
    """

    __implements__ = IPropertyAdapter

    def __init__(self, attribute_names):
        self._names = attribute_names

    def getProperties(self, obj):
        properties = []
        for name in self._names:
            attr = getattr(obj, name)
            typ = type_map.get(type(attr), 'string')
            properties.append({
                'id': name, 'value': attr, 'type': typ
                })
        return tuple(properties)

    def setProperties(self, obj, properties):
        names = self._names[:]
        for prop in properties:
            id = prop['id']
            if id in names:
                setattr(obj, id, prop['value'])
                names.remove(id)
        # remove the remaining
        for name in names:
            delattr(obj, name)


from OFS.DTMLMethod import DTMLMethod
propertyAdapterRegistry.registerForClass(
    DTMLMethod,
    AttributePropertyAdapter(['title'])
    )
