#
# RevisionManager
#
# Copyright (c) 2001 by beehive elektronische medien GmbH
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public
# License (MPL)
#

"""
$Id: RevisionManager.py,v 1.37 2003/05/22 15:00:44 philipp Exp $
"""

# standard python modules
import os
import shutil
import time
import struct
import posixpath
from StringIO import StringIO

# Zope modules
import Globals
from OFS import SimpleItem, Image
from AccessControl import Role, ClassSecurityInfo, getSecurityManager
from App.Dialogs import MessageDialog
from Acquisition import aq_parent, aq_base, aq_acquire, aq_inner
from DateTime import DateTime
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from ZPublisher.mapply import mapply

# product modules
import Permissions
from CVSWrapper import CVSWrapper, CVSError
from CVSObjects import CVSDirectory
from ObjectFactories import factoryRegistry
from Properties import propertyAdapterRegistry, propertySerializer
from util import is_binary, guess_meta_type

from interfaces.RevisionManager import IRevisionManager, IRevisionManagerZMI

# constants and global variables
FS_DIRNAME = "RevisionManager"
DATA_DIR = os.path.join(Globals.INSTANCE_HOME, "var", FS_DIRNAME)
CVSPASSFILE = os.path.join(DATA_DIR, 'cvspass')
BAD_IDS = ('acl_users', 'RevisionManager', 'Control_Panel')

# default status
default_status = {
    # standard CVS status information
    'filename':            '',
    'rpath':               '',
    'status':              '',
    'wrev':                '',
    'rrev':                '',
    'stag':                '',
    'sdate':               '',
    'sopt':                '',
    'tags':                {},
    # RevisionManager specific
    'zodb_last_modified':  0.0,
    'fs_last_modified':    0.0,
    'history_on_commit':   0.0,
    'have_props':          0,    # added
    'prop_status':         {},   # in
    'prop_last_modified':  0.0,  # v1.3
    }

class TreeObjectWrapper:
    """
    A wrapper around Zope objects to be more flexible when displaying
    them with ZopeTree.
    """

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, obj, path='', module_paths=[]):
        self.object = obj
        self.path = path
        self.id = obj.getId()
        self.folderish = obj.isPrincipiaFolderish
        self._module_paths = module_paths

    def getChildren(self):
        if not self.folderish:
            return []
        children = []
        objectIds = self.object.objectIds()
        objectIds.sort()
        for id in objectIds:
            # filter for bad IDs
            if id in BAD_IDS:
                continue
            path = posixpath.join(self.path, id)
            obj = getattr(self.object, id)
            children.append(TreeObjectWrapper(obj, path, self._module_paths))
        return children

    def supportsSourceMethods(self, wrapped_obj):
        if wrapped_obj.folderish:
            return 1
        base = aq_base(wrapped_obj.object)
        # object needs to support at least one method to retrieve its raw source
        # and one method to upload text to it
        return ((hasattr(base, 'manage_FTPget') or hasattr(base, 'EditableBody')
              or hasattr(base, 'document_src')  or hasattr(base, 'read'))
            and (hasattr(base, 'manage_upload') or hasattr(base, 'PUT')))

    def isNotInModule(self, wrapped_obj):
        if wrapped_obj.folderish:
            # even if a folder is already in the module, one of its children
            # could still not be in
            return len(wrapped_obj.getChildrenNotInModule()) > 0
        return wrapped_obj.path not in self._module_paths

    def getChildrenSupportingSourceMethods(self):
        children = self.getChildren()
        return filter(self.supportsSourceMethods, children)

    def getChildrenInModule(self):
        children = self.getChildren()
        return filter(lambda child, paths=self._module_paths:
                      child.path in paths, children)

    def getChildrenNotInModule(self):
        children = self.getChildren()
        return filter(self.isNotInModule, children)

class RevisionManager(SimpleItem.SimpleItem):
    """
    RevisionManager class
    """

    __implements__ = IRevisionManager, IRevisionManagerZMI

    # Zope specific stuff
    manage_options = (
        ( {'label': 'Status', 'action': 'rm_statusForm',
           'help': ('RevisionManager', 'status.stx') },
          {'label': 'Add objects', 'action': 'rm_addObjectsForm',
           'help': ('RevisionManager', 'addObjects.stx') },
          {'label': 'Map files to objects', 'action': 'rm_mapFSObjectsForm',
           'help': ('RevisionManager', 'mapFSObjects.stx') },
          {'label': 'Configuration', 'action': 'rm_configForm',
           'help': ('RevisionManager', 'config.stx')},
        )
        + SimpleItem.Item.manage_options
        + Role.RoleManager.manage_options
    )

    id = "RevisionManager"
    meta_type = "Revision Manager"
    security = ClassSecurityInfo()

    def __init__(self):
        # this title is temporary and will be replaced with the proper one
        # later on
        self.title = "not properly initialized"

        # private attributes
        self._settings = {}
        self._status = {}
        # create the data directory if it doesn't exist already
        if not os.path.isdir(DATA_DIR):
            os.mkdir(DATA_DIR)

    def _setId(self, id):
        # Don't permit rename
        if id != self.id:
            return MessageDialog(
                title="Invalid ID",
                message="It is not possible to change the ID of a " \
                        "RevisionManager object.",
                action="manage_workspace"
                )

    def _canCopy(self, op=0):
        # Don't permit copying/moving
        return 0

    # private auxiliary methods:

    def _clean_up(self):
        settings = self._settings
        if not settings.has_key('module'):
            return
	module_path = os.path.join(settings['work_dir'], settings['module'])
	if os.path.isdir(module_path):
	    try:
	        shutil.rmtree(module_path, 1)
	    except OSError:
		pass

    def _create_wrapper(self):
        settings = self._settings
        self._cvs_wrapper = CVSWrapper(**settings)

    def _get_wrapper(self):
        settings = self._settings
        if (not settings['use_zope_uid'] and
            settings['access_method'] != 'ext'):
            return self._cvs_wrapper
        kw = settings.copy()
        user = getSecurityManager().getUser()
        kw['user_name'] = user.getUserName()
        cvs_wrapper = CVSWrapper(**kw)
        return cvs_wrapper

    def _create_module_object(self):
        settings = self._settings
        work_dir = settings['work_dir']
        module = settings['module']
        module_path = posixpath.join(work_dir, module)
        self.module = CVSDirectory(module_path, "")

    def _create_fake_request(self):
        environ = os.environ.copy()
        environ['SERVER_NAME'] = 'localhost'
        environ['SERVER_PORT'] = ''
        response = HTTPResponse(stdout=StringIO())
        request = HTTPRequest(StringIO(''), environ, response)
        return request

    def _get_source(self, obj, request):
        """
        Get the source of an object
        """
        if (hasattr(aq_base(obj), 'data') and hasattr(obj.data, '__class__')
            and obj.data.__class__ is Image.Pdata):
            # yuck! we have a File instance with chunked data, let's stream it!
            result = ""
            data = obj.data
            while data is not None:
                result += data.data
                data = data.next
            return result

        kw = { 'REQUEST': request, 'RESPONSE': request.RESPONSE }
        
        if hasattr(obj, 'document_src'):
            result = mapply(obj.document_src, keyword=kw)
        elif hasattr(obj, 'read'):
            result = obj.read()
        elif hasattr(obj, 'EditableBody'):
            result = obj.EditableBody()
        elif hasattr(obj, 'manage_FTPget'):
            result = mapply(obj.manage_FTPget, keyword=kw)
        else:
            result = ""

        return str(result)

    def _put_source(self, obj, src):
        """
        Upload text to an object
        """
        request = self._create_fake_request()
        request.other['BODY'] = request.other['BODYFILE'] = src
        base = aq_base(obj)
        if hasattr(base, "PUT"):
            kw = { 'REQUEST': request, 'RESPONSE': request.RESPONSE }
            mapply(obj.PUT, keyword=kw)
        elif hasattr(base, "manage_upload"):
            obj.manage_upload(src)
        else:
            return 1  # error
        return 0

    def _escape_spaces(self, paths):
        return [path.replace(' ', '\ ') for path in paths]

    ############################################################################
    # Implementation methods for interface
    # interfaces.RevisionManager.IRevisionManager
    #

    security.declareProtected(Permissions.RMAdd, "getSecondStage")
    def getSecondStage(self, branch):
        try:
            self.checkout(tag=branch)
        except CVSError:
            pass
        
        if self.haveWorkingCopy():
            paths = self.getFilePaths()
            self.status(paths)
            # we're properly initialized
            self.title = self.id
            return 'rm_mapFSObjectsForm'
        return 'rm_importForm'

    # settings

    security.declareProtected(Permissions.RMManage, "getSettings")
    def getSettings(self):
        return self._settings.copy()

    security.declareProtected(Permissions.RMConfig, "setSettings")
    def setSettings(self, settings):
        self._settings.update(settings)
        self._p_changed = 1

    security.declareProtected(Permissions.RMConfig, "saveSettings")
    def saveSettings(self, mapping, module=None):
        settings = {
            'cvs_bin':         mapping.get("cvs_bin",         "cvs"),
            'access_method':   mapping.get("access_method",   "local"),
            'repository_path': mapping.get("repository_path", "/var/cvs"),
            'work_dir':        DATA_DIR,
            'server_addr':     mapping.get("server_addr",     ""),
            'user_name':       mapping.get("user_name",       ""),
            'passwd':          mapping.get("passwd",          ""),
            'use_zope_uid':    mapping.get("use_zope_uid",    0),
            'cvs_rsh':         mapping.get("cvs_rsh",         "rsh"),
            'comp_level':      mapping.get("comp_level",      "3"),
            'save_properties': mapping.get("save_properties", 1),
            'cvs_passfile':    CVSPASSFILE,
            'include_history': mapping.get('include_history', 0),
            'debug':           mapping.get("debug",           0)
            }
        if module is not None:
            settings['module'] = module
        self.setSettings(settings)
        if (not settings['use_zope_uid'] or
            settings['access_method'] == 'local'):
            self._create_wrapper()
        else:
            try:
                del self._cvs_wrapper
            # don't laugh, it really raises KeyError, not
            # AttributeError
            except KeyError:
                pass
        if module is not None:
            self._create_module_object()

    # lazy traverse

    security.declarePrivate("lazyTraverse")
    def lazyTraverse(self, path, default=None):
        if not hasattr(self, '_v_lazy_traverse'):
            self._v_lazy_traverse = {}

        if not self._v_lazy_traverse.has_key(path):
            obj = aq_parent(aq_inner(self))
            if path != "":
                obj = obj.restrictedTraverse(path, default)
            # set the shortcut
            self._v_lazy_traverse[path] = obj
            return obj

        # the shortcut is present, so use it
        return self._v_lazy_traverse[path]

    security.declarePrivate("setLazyTraverse")
    def setLazyTraverse(self, path, obj):
        # do nothing if lazy traverse has not been initialized
        if not hasattr(self, '_v_lazy_traverse'):
            self._v_lazy_traverse = {}
        self._v_lazy_traverse[path] = obj

    security.declarePrivate("clearLazyTraverse")
    def clearLazyTraverse(self):
        if hasattr(self, '_v_lazy_traverse'):
            del self._v_lazy_traverse

    # status

    security.declareProtected(Permissions.RMManage, "getStatus")
    def getStatus(self, path=None):
        if not path:
            return self._status.copy()
        return self._status[path]

    security.declareProtected(Permissions.RMManage, "setStatus")
    def setStatus(self, statusinfo):
        status = self._status
        for path, statusdict in statusinfo.items():
            # treat property files differently
            basename = posixpath.basename(path)
            if basename.startswith('.') and basename.endswith('.properties'):
                path = posixpath.join(posixpath.dirname(path), basename[1:-11])
                if not status.has_key(path):
                    status[path] = default_status.copy()
                new_prop_status = status[path].get('prop_status', {})
                new_prop_status.update(statusdict)
                status[path]['prop_status'] = new_prop_status
            elif not status.has_key(path):
                status[path] = statusdict
            else:
                # it is important that we use the update() method here
                # because the statusdict might only contain updates,
                # not a complete set of status information
                status[path].update(statusdict)
        self._p_changed = 1

    # cvs objects

    security.declareProtected(Permissions.RMManage, "getModuleFolder")
    def getModuleFolder(self):
        # we can't just do aq_parent(self) because we might have been
        # acquired. Use the following trick to retrieve the physical
        # parent
        return aq_parent(aq_inner(self))

    security.declareProtected(Permissions.RMManage, "getTreeWrappedModuleFolder")
    def getTreeWrappedModuleFolder(self):
        module_folder = aq_parent(aq_inner(self))
        module_paths = self.getPaths()
        return TreeObjectWrapper(module_folder, module_paths=module_paths)

    security.declareProtected(Permissions.RMManage, "getCVSModule")
    def getCVSModule(self):
        if not hasattr(self, 'module'):
            self._create_module_object()
        return self.module

    security.declareProtected(Permissions.RMManage, "traverseCVSObject")
    def traverseCVSObject(self, path):
        module = self.getCVSModule()
        return module.traverse(path)

    # paths

    security.declarePrivate("getPaths")
    def getPaths(self):
        return self._status.keys()

    security.declarePrivate("getAbsoluteFilePath")
    def getAbsoluteFilePath(self, path):
        settings = self._settings
        return os.path.join(settings['work_dir'], settings['module'], path)

    security.declarePrivate("getSubobjectsPaths")
    def getSubobjectsPaths(self, path, obj=None):
        if obj is None:
            obj = self.lazyTraverse(path)
        if obj is None:
            # someone has given us a fake path
            return []
        if not obj.isPrincipiaFolderish:
            return []
        paths = []
        for id, subobj in obj.objectItems():
            # filter for bad IDS
            if id in BAD_IDS:
                continue
            subpath = posixpath.join(path, id) 
            paths.append(subpath)
            # while we're at it, remember the object for lazy traversing
            self.setLazyTraverse(subpath, subobj)
            if obj.isPrincipiaFolderish:
                paths += self.getSubobjectsPaths(subpath, subobj)
        return paths

    security.declarePrivate("getFilePaths")
    def getFilePaths(self):
        module = self.getCVSModule()
        return module.getDirectoryListing(recursive=1)

    security.declarePrivate("addParentPaths")
    def addParentPaths(self, paths):
        all_paths = paths[:]
        for path in paths:
            segments = path.split("/")
            for i in range(1, len(segments)):
                dir_path = "/".join(segments[:i])
                if dir_path not in all_paths:
                    all_paths.append(dir_path)
        return all_paths

    security.declarePrivate("addChildPaths")
    def addChildPaths(self, paths):
        module_paths = self._status.keys()
        # shortcut
        if '' in paths:
            return module_paths

        all_paths = paths[:]
        for path in paths:
            sub_paths = [module_path for module_path in module_paths
                         if module_path.startswith(path+'/')]
            all_paths += sub_paths
        all_paths.sort()
        return all_paths

    security.declarePrivate("addParentAndChildPaths")
    def addParentAndChildPaths(self, paths):
        module_paths = self._status.keys()
        # shortcut
        if '' in paths:
            return module_paths

        all_paths = self.addParentPaths(paths)
        all_paths = self.addChildPaths(all_paths)
        all_paths.sort()
        return all_paths

    security.declarePrivate("addParentAndChildObjectPaths")
    def addParentAndChildObjectPaths(self, paths):
        all_paths = self.addParentPaths(paths)
        for path in paths:
            all_paths += self.getSubobjectsPaths(path)
        all_paths.sort()
        return all_paths

    security.declarePrivate("addPropertyPaths")
    def addPropertyPaths(self, paths):
        status = self._status
        all_paths = paths[:]
        for path in paths:
            if path and status.has_key(path) and status[path]['have_props']:
                dirname = posixpath.dirname(path)
                basename = posixpath.basename(path)
                all_paths.append(
                    posixpath.join(dirname, "." + basename + ".properties")
                    )
        return all_paths

    security.declareProtected(Permissions.RMManage, "getFSFilesNotInZODB")
    def getFSFilesNotInZODB(self):
        module_path = self.getCVSModule().absolute_path
        all_fspaths = self.getFilePaths()
        notinzodb = []
        module_folder = aq_parent(aq_inner(self))
        for path in all_fspaths:
            # use real restrictedTraverse() here
            obj = module_folder.restrictedTraverse(path, None)
            if obj is None:
                abs_path = os.path.join(module_path, path)
                if not os.path.isfile(abs_path):
                    continue
                fp = open(abs_path)
                data = fp.read()
                fp.close()
                notinzodb.append(
                    {'path': path, 'meta_type': guess_meta_type(data)}
                    )
        return notinzodb

    # history

    security.declarePrivate("getHistory")
    def getHistory(self, path):
        status = self.getStatus(path)
        last_history = status.get('history_on_commit', 0)
        obj = self.lazyTraverse(path)
        if obj is None:
            return []

        start = 0
        batchsize = end = 32
        history = []
        foundit = 0
        # the result of the history function can be pretty long, so we
        # batch it
        while 1:
            batch = obj._p_jar.db().history(obj._p_oid, None, end)
            batch = batch[start:]
            if not batch:
                # we're already at the end of the batch
                break
            for hist in batch:
                # only return history information that has not been
                # comitted yet
                if hist['time'] > last_history:
                    history.append(hist)
                else:
                    foundit = 1
                    break
            if foundit:
                break
            start += batchsize
            end += batchsize
        history.reverse()
        return history

    security.declarePrivate("getHistoryAsText")
    def getHistoryAsText(self, history):
        output = ""
        for hist in history:
            datetime = DateTime(hist['time'])
            output += "  %s %s\t%s\n" % (datetime.ISO(),
                                         hist['user_name'],
                                         hist['description'])
        return output

    # other

    security.declareProtected(Permissions.RMManage, "getSupportedMetaTypes")
    def getSupportedFactoryMetaTypes(self):
        meta_types = factoryRegistry.getSupportedMetaTypes()
        meta_types = list(meta_types)
        meta_types.sort()
        return tuple(meta_types)

    security.declareProtected(Permissions.RMManage, "upgrade")
    def upgrade(self):
        settings = self._settings

        # 1.0 to 1.1
        if not settings.has_key('work_dir'):
            settings['work_dir'] = DATA_DIR
        
        # 1.1 to 1.2
        if not settings.has_key('use_zope_uid'):
            settings['use_zope_uid'] = 0
        if not settings.has_key('include_history'):
            settings['include_history'] = 0
        if not hasattr(self, '_status'):
            status = self._status = {}
            for path in self.getFilePaths():
                status[path] = default_status.copy()
                status[path]['prop_status'] = {}
            self.status([''])

        # 1.1 to 1.3
        status = self._status
        if not settings.has_key('save_properties'):
            settings['save_properties'] = 0
        for statusdict in status.values():
            d = {}
            statusdict.update({
                'have_props':          0,
                'prop_status':         d,
                'prop_last_modified':  0.0,
                })

    security.declareProtected(Permissions.RMManage, "upgradeToProperties")
    def upgradeToProperties(self):
        settings = self._settings
        status = self._status

        settings['save_properties'] = 1

        # flush the cache
        for statusdict in status.values():
            statusdict['zodb_last_modified'] = 0.0

        # write everything to the file system again to be sure
        paths = self.getPaths()
        self.writeToFS(paths)

        # add all .properties files to the cvs and commit them
        prop_paths = []
        for path in paths:
            if status[path]['have_props']:
                dirname = posixpath.dirname(path)
                basename = posixpath.basename(path)
                path = posixpath.join(dirname, "." + basename + ".properties")
                prop_paths.append(path)
        prop_paths = self._escape_spaces(prop_paths)
        cvs = self._get_wrapper()
        cvs.add(" ".join(prop_paths))
        cvs.commit("Added properties by RevisionManager", " ".join(prop_paths))

        # read the stuff back in and update status
        self.readFromFS(paths)
        self.status([''])

        self._p_changed = 1

    #
    # standard cvs actions
    #

    security.declareProtected(Permissions.RMManage, "importModule")
    def importModule(self, vendor_tag, release_tag, log_message, paths):
        module = self.getCVSModule()
        # create the import directory
        #XXX not portable
        import_dir = "/tmp/RevisionManager-" + module.filename
        os.mkdir(import_dir)

        # write objects to the file system
        paths = self.addParentAndChildObjectPaths(paths)
        self.writeToFS(paths, import_dir)

        # import the module
        cvs = self._get_wrapper()
        try:
            cvs.importModule(vendor_tag, release_tag, log_message,
                             import_dir=import_dir)
        finally:
            # delete the import directory
            shutil.rmtree(import_dir, 1)

        # do a checkout right away
        self.checkout()
        self.readFromFS(paths)
        self.status([''])

        # now we can say that we're properly initialized, so replace the title
        self.title = self.id
        self.clearLazyTraverse()

    security.declareProtected(Permissions.RMUpdate, "checkout")
    def checkout(self, tag=''):
        cvs = self._get_wrapper()
        cvs.checkout(tag=tag, prune_empty=1)

    security.declareProtected(Permissions.RMUpdate, "forceCheckout")
    def forceCheckout(self, paths):
        # force checkout the whole module
        if '' in paths:
            # first delete the whole working copy
            settings = self._settings
            module_path = os.path.join(settings['work_dir'], settings['module'])
            shutil.rmtree(module_path, 1)
            # then check it out again
            self.checkout()
            # now read it to the objects
            all_paths = self.addParentAndChildPaths(paths)
            self.readFromFS(all_paths)
            self.status(paths)
            self.clearLazyTraverse()
            return

        # only selected files
        paths = self.getLocallyModifiedFiles(paths)
        if not paths:
            self.clearLazyTraverse()
            return

        settings = self._settings
        if settings['save_properties']:
            paths = self.addPropertyPaths(paths)

        # first delete each given path
        for path in paths:
            absolute_path = self.getAbsoluteFilePath(path)
            os.unlink(absolute_path)

        # then do a cvs update
        escaped_paths = self._escape_spaces(paths)
        cvs = self._get_wrapper()
        cvs_output = cvs.update(" ".join(escaped_paths), prune_empty=1,
                                new_dirs=1)

        # check for properties in cvs update
        if settings['save_properties']:
            for path, cvsstatus in cvs_output.items():
                basename = posixpath.basename(path)
                if basename.startswith('.') and basename.endswith('.properties'):
                    realpath = posixpath.join(posixpath.dirname(path), basename[1:-11])
                    cvs_output[realpath] = ''
                    del cvs_output[path]

        # now read it to the objects and update the status
        updated_paths = cvs_output.keys()
        self.readFromFS(updated_paths)
        self.status(updated_paths)
        self.clearLazyTraverse()

    security.declareProtected(Permissions.RMUpdate, "update")
    def update(self, paths, tag='', reset_sticky=0):
        if not paths:
            return []

        # write objects to FS
        all_paths = self.addParentAndChildPaths(paths)
        self.writeToFS(all_paths)

        settings = self._settings
        if settings['save_properties']:
            paths = self.addPropertyPaths(paths)
        escaped_paths = self._escape_spaces(paths)

        # do the CVS update
        cvs = self._get_wrapper()
        cvs_output = cvs.update(" ".join(escaped_paths), tag=tag,
                                prune_empty=1, new_dirs=1,
                                reset_sticky=reset_sticky)

        # check for properties in cvs update
        if settings['save_properties']:
            for path, cvsstatus in cvs_output.items():
                basename = posixpath.basename(path)
                if basename.startswith('.') and basename.endswith('.properties'):
                    realpath = posixpath.join(posixpath.dirname(path), basename[1:-11])
                    if cvs_output.has_key(realpath):
                        cvs_output[realpath] += "; Properties: " + cvsstatus
                    else:
                        cvs_output[realpath] = "Properties: " + cvsstatus
                    del cvs_output[path]

        status = self._status
        notinzodb = []
        updated_paths = cvs_output.keys()
        if updated_paths:
            # read the files into the objects again
            notinzodb, prop_errors = self.readFromFS(updated_paths)

        # update the status in CVSObjects
        if tag or reset_sticky:
            self.status(all_paths)
        elif updated_paths:
            self.status(updated_paths)

        return_list = []
        internal_paths = self.getPaths()
        for path, cvsstatus in cvs_output.items():
            # add to internal status dictionary if the file is new
            if path not in internal_paths:
                status[path] = default_status.copy()
                status[path]['prop_status'] = {}

            # remove from internal status dictionary if the file has
            # been removed from the repository
            if cvsstatus == "Removed from repository":
                del status[path]

            # prepare the return output
            extra = ""
            if path in notinzodb:
                extra = "Not in ZODB"
            if path in prop_errors:
                extra = "Error when setting properties."
            dict = {'path':   path,
                    'status': cvsstatus,
                    'extra':  extra }
            return_list.append(dict)

        self.clearLazyTraverse()
        self._p_changed = 1
        return return_list

    security.declareProtected(Permissions.RMCommit, "commit")
    def commit(self, log_message, paths):
        settings = self._settings
        status = self._status
        # we assume that this already contains the complete list of
        # modified files so simply write those objects to FS
        self.writeToFS(paths)

        # commit in cvs
        cvs = self._get_wrapper()
        if settings['include_history']:
            for path in paths:
                history = self.getHistory(path)
                assert len(history)
                # remember the time of the last history item
                status = self.getStatus(path)
                status['history_on_commit'] = history[-1]['time']
                # append the textual representation of the history to
                # the log message
                histtext = self.getHistoryAsText(history)
                log_hist = "%s\n\nZODB history since last commit:\n%s" % \
                           (log_message, histtext)
                path = path.replace(' ', '\ ')

                if settings['save_properties'] and status['have_props']:
                    dirname = posixpath.dirname(path)
                    basename = posixpath.basename(path)
                    path = path + ' ' + posixpath.join(dirname, '.' + basename + '.properties')

                # now commit it
                # Warning: this will be slow since we fire up a commit
                # for each object
                cvs.commit(log_hist, path)
            self._p_changed = 1
        else:
            if settings['save_properties']:
                escaped_paths = self._escape_spaces(self.addPropertyPaths(paths))
            else:
                escaped_paths = self._escape_spaces(paths)
            output = cvs.commit(log_message, " ".join(escaped_paths))

        # write the FS file contents back again in case it contained $...$
        self.readFromFS(paths)
        self.status(paths)

    security.declareProtected(Permissions.RMAddRemove, "add")
    def add(self, paths):
        status = self._status
        save_properties = self._settings['save_properties']

        all_paths = self.addParentAndChildObjectPaths(paths)
        module_paths = self.getPaths()
        self.writeToFS(all_paths)

        file_paths = []
        binary_file_paths = []
        dir_paths = []
        for path in all_paths:
            if path in module_paths:
                continue
            
            obj = self.lazyTraverse(path)
            if obj.isPrincipiaFolderish:
                dir_paths.append(path)
            else:
                # check for binariness
                filename = self.getAbsoluteFilePath(path)
                fp = open(filename)
                bin = is_binary(fp)
                fp.close()
                
                # escape spaces
                if bin:
                    binary_file_paths.append(path)
                else:
                    file_paths.append(path)

        cvs = self._get_wrapper()
        # first add the directories. They don't need a commit.
        if dir_paths:
            cvs.add(" ".join(self._escape_spaces(dir_paths)))

        dir_property_paths = []
        if save_properties:
            # add property files for directories
            for path in dir_paths:
                if status[path]['have_props']:
                    dirname = posixpath.dirname(path)
                    basename = posixpath.basename(path)
                    prop_path = posixpath.join(dirname, '.' + basename + '.properties')
                    dir_property_paths.append(prop_path)

            # add property files
            file_paths = self.addPropertyPaths(file_paths) + dir_property_paths
            binary_file_paths = self.addPropertyPaths(binary_file_paths)

        # concatenate paths and escape spaces
        file_paths = " ".join(self._escape_spaces(file_paths))
        binary_file_paths = " ".join(self._escape_spaces(binary_file_paths))

        # now add regular and binary files and commit them.
        if file_paths:
            cvs.add(file_paths)
        if binary_file_paths:
            cvs.add(binary_file_paths, binary=1)
        cvs.commit("File added to repository by RevisionManager.",
                   file_paths + " " + binary_file_paths)

        # update status
        self.status(paths)
        self.clearLazyTraverse()

    security.declareProtected(Permissions.RMAddRemove, "remove")
    def remove(self, paths):
        status = self._status
        # remember all directories in paths. They will be deleted later
        dir_filenames = []
        for path in paths:
            filename = self.getAbsoluteFilePath(path)
            if os.path.isdir(filename):
                dir_filenames.append(filename)

        # expand the list to include all parent and child paths
        paths = self.addChildPaths(paths)

        # add property paths if necessary
        if self._settings['save_properties']:
            paths_with_props = self.addPropertyPaths(paths)
            # remove from internal status dict
            for path in paths:
                if status.has_key(path):
                    del status[path]
            paths = paths_with_props
        else:
            # remove from internal status dict
            for path in paths:
                del self._status[path]

        file_paths = []
        dir_paths = []
        for path in paths:
            filename = self.getAbsoluteFilePath(path)
            escaped_path = path.replace(' ', '\ ')
            if os.path.isdir(filename):
                # remember directories
                dir_paths.append(escaped_path)
            elif os.path.isfile(filename):
                # delete regular files right away
                os.unlink(filename)
                file_paths.append(escaped_path)

        # remove regular files from cvs
        cvs = self._get_wrapper()
        if file_paths:
            file_paths = " ".join(file_paths)
            cvs.remove(file_paths)
            # file won't be removed from the repository unless they are committed
            cvs.commit("Removed file by RevisionManager.", file_paths)

        # remove directories from cvs
        if dir_paths:
            cvs.remove(" ".join(dir_paths))

        # finally, delete directories recursively
        for directory in dir_filenames:
            shutil.rmtree(directory, 1)

        self._p_changed = 1
        self.clearLazyTraverse()

    security.declareProtected(Permissions.RMManage, "status")
    def status(self, paths):
        if not paths:
            return {}

        # write objects to FS
        all_paths = self.addParentAndChildPaths(paths)
        self.writeToFS(all_paths)

        if self._settings['save_properties']:
            paths = self.addPropertyPaths(paths)
        paths = self._escape_spaces(paths)

        # update status info from cvs
        cvs = self._get_wrapper()
        status = cvs.status(" ".join(paths), verbose=1)
        self.setStatus(status)
        self.clearLazyTraverse()
        return status

    security.declareProtected(Permissions.RMManage, "log")
    def log(self, paths):
        if not paths:
            return {}

        # write objects to FS
        all_paths = self.addParentAndChildPaths(paths)
        self.writeToFS(all_paths)

        if self._settings['save_properties']:
            paths = self.addPropertyPaths(paths)
        paths = self._escape_spaces(paths)

        # retrieve cvs log information
        cvs = self._get_wrapper()
        self.clearLazyTraverse()
        return cvs.log(" ".join(paths))

    security.declareProtected(Permissions.RMManage, "diff")
    def diff(self, paths, rev1='', rev2=''):
        if not paths:
            return ''

        # write objects to FS
        all_paths = self.addParentAndChildPaths(paths)
        self.writeToFS(all_paths)

        if self._settings['save_properties']:
            paths = self.addPropertyPaths(paths)
        paths = self._escape_spaces(paths)

        # retrieve cvs diff information
        cvs = self._get_wrapper()
        self.clearLazyTraverse()
        return cvs.diff(" ".join(paths), rev1, rev2)

    security.declareProtected(Permissions.RMManage, "tag")
    def tag(self, tag, paths, branch=0):
        if self._settings['save_properties']:
            paths = self.addPropertyPaths(paths)
        paths = self._escape_spaces(paths)

        cvs = self._get_wrapper()
        return cvs.tag(tag, " ".join(paths), branch=branch)

    security.declareProtected(Permissions.RMManage, "merge")
    def merge(self, path, rev1, rev2=""):
        if not rev2:
            rev2 = rev1
            status = self.status([path])
            rev1 = status[path]['wrev']
        escaped_path = path.replace(' ', '\ ')
        cvs = self._get_wrapper()
        if rev1 == rev2:
            filename = self.getAbsoluteFilePath(path)
            os.unlink(filename)
            cvs.update(escaped_path)
        else:
            cvs.merge(rev1, rev2, escaped_path)
        self.readFromFS([path])
        self.status([path])

    security.declareProtected(Permissions.RMManage, "getModifiedFiles")
    def getLocallyModifiedFiles(self, paths=[]):
        if not paths:
            paths = ['']
            
        # write objects to FS
        all_paths = self.addParentAndChildPaths(paths)
        self.writeToFS(all_paths)

        if self._settings['save_properties']:
            paths = self.addPropertyPaths(paths)
        paths = self._escape_spaces(paths)

        # retrieve information about locally modified files
        cvs = self._get_wrapper()
        self.clearLazyTraverse()
        locally_modified = cvs.getLocallyModifiedFiles(' '.join(paths))
        for path in locally_modified[:]:
            basename = posixpath.basename(path)
            if basename.startswith('.') and basename.endswith('.properties'):
                locally_modified.remove(path)
                real_path = posixpath.join(posixpath.dirname(path), basename[1:-11])
                if real_path not in locally_modified:
                    locally_modified.append(real_path)
        locally_modified.sort()
        return locally_modified

    # file system interaction

    security.declarePrivate("writeToFS")
    def writeToFS(self, paths, module_path=None):
        settings = self._settings
        status = self._status
        
        if module_path is None:
            module_path = os.path.join(settings['work_dir'],
                                       settings['module'])

        # create module directory if not present already
        if settings['module'] not in os.listdir(settings['work_dir']):
            try:
                os.mkdir(module_path)
            except OSError:
                pass

        error_paths = []
        fake_request = self._create_fake_request()
        # prepare a string representation of paths because we might
        # want to search it later
        _concat_paths = ' '.join(paths)
        for path in paths:
            # traverse to the current object
            obj = self.lazyTraverse(path)
            if obj is None:
                error_paths.append(path)
                continue

            # ignore empty folderish objects
            if obj.isPrincipiaFolderish:
                # check whether paths contains subpaths for that
                # object. Use the find method of the string
                # representation instead of some custom loop because
                # it's faster (implemented in C)
                #
                # also, this keeps it from writing the properties of
                # the module root folder
                if _concat_paths.find(' ' + path + '/') == -1:
                    continue

            abs_path = os.path.join(module_path, path)

            # acquire the status for the current object
            if not status.has_key(path):
                obj_status = status[path] = default_status.copy()
                obj_status['prop_status'] = {}
            else:
                obj_status = status[path]

            # save properties if necessary
            if settings['save_properties']:
                self.writePropsToFS(obj, abs_path, obj_status)

            # handle folderish objects
            if obj.isPrincipiaFolderish:
                if not os.path.isdir(abs_path):
                    os.mkdir(abs_path)
                continue

            # don't write if it wasn't modified
            obj_last_modified = float(obj.bobobase_modification_time())
            if obj_last_modified == obj_status.get('zodb_last_modified', 0.0):
                continue

            obj_status['zodb_last_modified'] = obj_last_modified
            src = self._get_source(obj, fake_request)

            # open, write to file and close it again
            fp = open(abs_path, "wb")
            fp.write(src)
            fp.close()
            obj_status['fs_last_modified'] = float(os.stat(abs_path)[9])

        # Set dirty flag. It's better to set it once here than setting it many
        # times in the above for loop.
        self._p_changed = 1
        return error_paths

    security.declarePrivate("readFromFS")
    def readFromFS(self, paths):
        settings = self._settings
        status = self._status
        
        notinzodb = []
        prop_errors = []
        module_path = os.path.join(settings['work_dir'], settings['module'])
        fake_request = self._create_fake_request()
        for path in paths:
            # traverse to the object
            obj = self.lazyTraverse(path)
            if obj is None:
                # remember if the corresponding object could not be found
                notinzodb.append(path)
                continue

            abs_path = os.path.join(module_path, path)

            # read properties
            if settings['save_properties']:
                if status.has_key(path):
                    try:
                        self.readPropsFromFS(obj, abs_path, status[path])
                        if settings['include_history']:
                            self.rememberLastHistory(path, status[path])
                    except:
                        prop_errors.append(path)

            # do nothing with folderish objects
            if obj.isPrincipiaFolderish:
                continue

            # abort if the file does not exist
            if not os.path.exists(abs_path):
                continue

            # don't write if it wasn't modified
            obj_status = status[path]
            file_last_modified = float(os.stat(abs_path)[9])
            if file_last_modified == obj_status.get('fs_last_modified', 0.0):
                continue

            obj_status['fs_last_modified'] = file_last_modified
            fp = open(abs_path, "rb")
            src = fp.read()
            fp.close()
            # only modify if necessary
            if self._get_source(obj, fake_request) != src:
                self._put_source(obj, src)
                # remember that we have changed the objects everywhere
                # necessary
                obj_status['zodb_last_modified'] = float(obj.bobobase_modification_time())
                if settings['include_history']:
                    self.rememberLastHistory(path, obj_status)

        self._p_changed = 1
        return notinzodb, prop_errors

    security.declarePrivate("rememberLastHistory")
    def rememberLastHistory(self, path, obj_status):
        # make sure, the previous changes are committed
        # XXX there is one problem with this, though. If an other part
        # of RevisionManager raises an exception *after* this method
        # has been called, changes will only be rolled back up to this
        # point.
        get_transaction().commit()
        history = self.getHistory(path)
        assert len(history)
        obj_status['history_on_commit'] = history[-1]['time']

    # properties

    security.declarePrivate("writePropsToFS")
    def writePropsToFS(self, obj, abs_path, obj_status):
        # try to find an adapter; if it fails abort
        adapter = propertyAdapterRegistry.getAdapter(obj)
        if adapter is None:
            obj_status['have_props'] = 0
            return

        # don't write if it wasn't modified
        obj_last_modified = float(obj.bobobase_modification_time())
        if obj_last_modified == obj_status.get('zodb_last_modified', 0.0):
            return

        properties = adapter.getProperties(obj)
        serial = propertySerializer.serialize(properties)
        basename = os.path.basename(abs_path)
        dirname = os.path.dirname(abs_path)

        # ignore the module folder
        if dirname == self._settings['work_dir']:
            return

        abs_path = os.path.join(dirname, "." + basename + '.properties')
        fp = open(abs_path, "wb")
        fp.write(serial)
        fp.close()

        obj_status['have_props'] = 1
        obj_status['prop_last_modified'] = float(os.stat(abs_path)[9])

    security.declarePrivate("readPropsFromFS")
    def readPropsFromFS(self, obj, abs_path, obj_status):
        if not obj_status['have_props']:
            return

        # try to find an adapter; if it fails, abort
        adapter = propertyAdapterRegistry.getAdapter(obj)
        if adapter is None:
            obj_status['have_props'] = 0
            return

        basename = os.path.basename(abs_path)
        dirname = os.path.dirname(abs_path)
        # ignore the module folder
        if dirname == self._settings['work_dir']:
            return
        abs_path = os.path.join(dirname, "." + basename + '.properties')

        prop_last_modified = float(os.stat(abs_path)[9])
        if prop_last_modified == obj_status.get('prop_last_modified', 0.0):
            return

        fp = open(abs_path, "rb")
        serial = fp.read()
        fp.close()
        properties = propertySerializer.deserialize(serial)

        # only set the properties if necessary
        if properties != adapter.getProperties(obj):
            adapter.setProperties(obj, properties)
        obj_status['zodb_last_modified'] = float(obj.bobobase_modification_time())

    security.declareProtected(Permissions.RMManage, "haveWorkingCopy")
    def haveWorkingCopy(self):
        if not self._settings.has_key('module'):
            return 0
        settings = self._settings
        module_path = os.path.join(settings['work_dir'], settings['module'])
        if not os.path.isdir(module_path):
            return 0
        cvsdir = os.path.join(module_path, 'CVS')
        if os.path.isdir(cvsdir):
            list = os.listdir(cvsdir)
            if 'Entries' in list and 'Repository' in list and 'Root' in list:
                return 1
        return 0

    security.declareProtected(Permissions.RMManage, 'addFSFileObject')
    def addFSFileObject(self, path, meta_type):
        segments = path.split("/")
        # first make sure that the directory part of the path exists
        # if it doesn't, create as many folders as necessary
        folder = aq_parent(aq_inner(self))
        for id in segments[:-1]:
            subfolder = getattr(folder, id, None)
            if subfolder is None:
                folder.manage_addFolder(id)
                subfolder = getattr(folder, id)
            folder = subfolder

        # now add the object
        # first the the appropriate factory
        factory = factoryRegistry.getFactory(meta_type)
        id = segments[-1]
        abs_filename = self.getAbsoluteFilePath(path)
        # read the file contents
        data = open(abs_filename, "rb").read()

        if not self._settings['save_properties']:
            try:
                obj = factory.createObject(id, id, data)
            except:
                return 1
            folder._setObject(id, obj)
            return 0

        basename = os.path.basename(abs_filename)
        dirname = os.path.dirname(abs_filename)
        prop_path = os.path.join(dirname, "." + basename + '.properties')

        properties = ()
        title = id
        props = {}
        # it could be that there is no properties file, so we
        # better check
        if os.path.isfile(prop_path):
            # read the properties file and deserialize the data
            fp = open(prop_path, "rb")
            serial = fp.read()
            fp.close()
            properties = propertySerializer.deserialize(serial)

            for propdict in properties:
                if propdict['id'] == 'title':
                    title = propdict['value']
                else:
                    props[propdict['id']] = propdict['value']

        # create the object
        try:
            obj = factory.createObject(id, title, data, **props)
        except:
            return 1

        # write properties if available and possible
        if properties:
            adapter = propertyAdapterRegistry.getAdapter(obj)
            if adapter is not None:
                obj_status = self.getStatus(path)
                obj_status['have_props'] = 1
                if properties != adapter.getProperties(obj):
                    try:
                        adapter.setProperties(obj, properties)
                    except:
                        return 1

        folder._setObject(id, obj)
        return 0

    #
    ############################################################################
    # Overwritten SimpleItem methods
    #

    def manage_beforeDelete(self, item, container):
        self._clean_up()

    ############################################################################
    # Implementation methods for interface
    # interfaces.RevisionManager.IRevisionManagerZMI

    # management views
    security.declareProtected(
        Permissions.RMManage, "rm_statusForm", "rm_addObjectsForm",
        "rm_configForm", "rm_importForm", "rm_mapFSObjectsForm",
        "rm_detailsForm", "rm_logForm", "rm_diffForm", "rm_commitForm",
        "rm_updateForm"
        )
    rm_statusForm       = PageTemplateFile("www/status.pt", globals())
    rm_addObjectsForm   = PageTemplateFile("www/addObjects.pt", globals())
    rm_configForm       = PageTemplateFile("www/config.pt", globals())
    rm_importForm       = PageTemplateFile("www/import.pt", globals())
    rm_mapFSObjectsForm = PageTemplateFile("www/mapFSObjects.pt", globals())
    rm_detailsForm      = PageTemplateFile("www/details.pt", globals())
    rm_logForm          = PageTemplateFile("www/log.pt", globals())
    rm_diffForm         = PageTemplateFile("www/diff.pt", globals())
    rm_commitForm       = PageTemplateFile("www/commit.pt", globals())
    rm_updateForm       = PageTemplateFile("www/update.pt", globals())

    security.declareProtected(Permissions.RMAddRemove, "rm_import")
    def rm_import(self, REQUEST, RESPONSE, vendor_tag, release_tag,
                  log_message, paths=[]):
        """
        Properly import selected files
        """
        if not paths:
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to import. "\
                        "Please select one or more files or directories.",
                action="rm_importForm"
                )
        
        module = self.getCVSModule()        
        vendor_tag = vendor_tag or module.filename.upper()
        release_tag = vendor_tag or module.filename.upper() + "1"
        log_message = log_message or "RevisionManager initial import"
        if vendor_tag == release_tag:
            release_tag += "1"

        self.importModule(vendor_tag, release_tag, log_message, paths)

        message = "Module '%s' imported" % module.filename
        REQUEST['manage_tabs_message'] = message
        return self.rm_statusForm(self, REQUEST)

    security.declareProtected(Permissions.RMUpdate, "rm_forceCheckout")
    def rm_forceCheckout(self, REQUEST, RESPONSE, paths=[]):
        """
        Force checkout
        """
        if not len(paths):
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to force a checkout on." \
                        "Please select one or more files or directories.",
                action="rm_statusForm"
                )
        self.forceCheckout(paths)
        RESPONSE.redirect(self.absolute_url() + "/rm_statusForm")

    security.declareProtected(Permissions.RMManage, "rm_updateStatus")
    def rm_updateStatus(self, REQUEST, RESPONSE, paths=[]):
        """
        Update status information.
        """
        if not paths:
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to update its status. " \
                        "Please select one or more files or directories.",
                action="rm_statusForm"
                )
        self.status(paths)

        message = "Updated status information"
        REQUEST['manage_tabs_message'] = message
        return self.rm_statusForm(self, REQUEST)

    security.declareProtected(Permissions.RMAddRemove, "rm_mapFSObjects")
    def rm_mapFSObjects(self, REQUEST, RESPONSE, items=[]):
        """
        Map file system files to objects.
        """
        failed = []
        for item in items:
            if item.get('add', 0):
                path = item['path']
                rt = self.addFSFileObject(path, item['meta_type'])
                if rt:
                    failed.append(path)
        if failed:
            message = "Object creation failed for: %s" % ' '.join(failed)
            REQUEST['manage_tabs_message'] = message
        return self.rm_statusForm(self, REQUEST)

    security.declareProtected(Permissions.RMConfig, "rm_config")
    def rm_config(self, REQUEST, RESPONSE):
        """
        Save configuration settings.
        """
        # only take 'module' from REQUEST if settings doesn't already have it
        module = self._settings.get("module", REQUEST.get("module"))
        REQUEST.set("module", module)
        self.saveSettings(REQUEST)
        
        message = "Saved changes"
        REQUEST['manage_tabs_message'] = message
        return self.rm_configForm(self, REQUEST)

    security.declareProtected(Permissions.RMAddRemove, "rm_addObjects")
    def rm_addObjects(self, REQUEST, RESPONSE, paths=[]):
        """
        Add objects to the repository.
        """
        if not paths:
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to add. " \
                        "Please select one or more files or directories.",
                action="rm_addObjectsForm"
                )
        self.add(paths)
        RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMAddRemove, "rm_removeObjects")
    def rm_removeObjects(self, REQUEST, RESPONSE, paths=[]):
        """
        Remove objects from RevisionManager's list and delete their
        equivalents in the CVS repository
        """
        if not paths:
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to remove. " \
                        "Please select one or more files or directories.",
                action="rm_statusForm"
                )
        self.remove(paths)
        RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMCommit, "rm_commit")
    def rm_commit(self, REQUEST, RESPONSE, log_message, paths=[]):
        """
        Commit changes
        """
        if not len(paths):
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to commit. "\
                        "Please select one or more files or directories.",
                action="rm_statusForm"
                )
        self.commit(log_message, paths)
        RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMCommit, "rm_tag")
    def rm_tag(self, REQUEST, RESPONSE, tag, paths=[]):
        """
        Set tag on file(s)
        """
        if not paths:
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to tag. " \
                        "Please select one or more files or directories.",
                action="rm_statusForm"
                )
        try:
            self.tag(tag, paths)
        except CVSError, e:
            return MessageDialog(title="Bad tag!", message=str(e),
                                 action="rm_statusForm")
        
        RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMCommit, "rm_branch")
    def rm_branch(self, REQUEST, RESPONSE, tag, paths=[]):
        """
        Create a new branch
        """
        if not paths:
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything for creating a new " \
                        "branch. Please select one or more files or " \
                        "directories.",
                action="rm_statusForm"
                )
        try:
            self.tag(tag, paths, branch=1)
        except CVSError, e:
            return MessageDialog(title="Bad tag!", message=str(e),
                                 action="rm_statusForm")
        RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMUpdate, "rm_merge")
    def rm_merge(self, REQUEST, RESPONSE, path, revision):
        """
        Overwrite an object's data with its data from an older revision
        """
        self.merge(path, revision)
        RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    #
    ############################################################################

#
# object constructors
#

Globals.InitializeClass(RevisionManager)

def manage_addRevisionManager(self, REQUEST, RESPONSE, module, branch=''):
    """
    Create an instance of RevisionManager
    """
    # prevent instances in folders whose parents already contain an instance
    # (since it may lead to data inconsistency)
    try:
        rm = aq_acquire(self, "RevisionManager")
    except AttributeError:
        rm = None
    if rm is not None:
        return MessageDialog(
	    title="Detected a RevisionManager instance",
	    message="There already is a RevisionManager instance in this " \
                    "folder or one of its parent folders.",
	    action="manage_workspace"
	   )

    rm = RevisionManager()
    self.Destination()._setObject(rm.id, rm)
    rm.saveSettings(REQUEST.form, module)

    # get the instance acquisition wrapped and fire up second stage
    rm = getattr(self, rm.id)
    second_stage = rm.getSecondStage(branch)
    RESPONSE.redirect(rm.absolute_url()+'/'+second_stage)

manage_addForm = PageTemplateFile("www/add.pt", globals())
